@@ -1,6 +1,6 @@
 # This file is part of Patsy
 # Copyright (C) 2011-2013 Nathaniel Smith <njs@pobox.com>
-# See file COPYING for license information.
+# See file LICENSE.txt for license information.
 
 # Some generic utilities.
 
@@ -9,12 +9,25 @@ __all__ = ["atleast_2d_column_default", "uniqueify_list",
            "repr_pretty_delegate", "repr_pretty_impl",
            "SortAnythingKey", "safe_scalar_isnan", "safe_isnan",
            "iterable",
+           "have_pandas",
+           "have_pandas_categorical",
+           "have_pandas_categorical_dtype",
+           "pandas_Categorical_from_codes",
+           "pandas_Categorical_categories",
+           "pandas_Categorical_codes",
+           "safe_is_pandas_categorical_dtype",
+           "safe_is_pandas_categorical",
+           "safe_issubdtype",
+           "no_pickling",
+           "assert_no_pickling",
+           "safe_string_eq",
            ]
 
 import sys
 import numpy as np
-from cStringIO import StringIO
-from compat import optional_dep_ok
+import six
+from six.moves import cStringIO as StringIO
+from .compat import optional_dep_ok
 
 try:
     import pandas
@@ -27,6 +40,20 @@ else:
 # Can drop this guard whenever we drop support for such older versions of
 # pandas.
 have_pandas_categorical = (have_pandas and hasattr(pandas, "Categorical"))
+if not have_pandas:
+    have_pandas_categorical_dtype = False
+    _pandas_is_categorical_dtype = None
+else:
+    if hasattr(pandas, "api"):
+        # This is available starting in pandas v0.19.0
+        have_pandas_categorical_dtype = True
+        _pandas_is_categorical_dtype = pandas.api.types.is_categorical_dtype
+    else:
+        # This is needed for pandas v0.18.0 and earlier
+        _pandas_is_categorical_dtype = getattr(pandas.core.common,
+                                               "is_categorical_dtype", None)
+        have_pandas_categorical_dtype = (_pandas_is_categorical_dtype
+                                         is not None)
 
 # Passes through Series and DataFrames, call np.asarray() on everything else
 def asarray_or_pandas(a, copy=False, dtype=None, subok=False):
@@ -82,15 +109,15 @@ def test_asarray_or_pandas():
                               columns=["A", "B", "C"],
                               index=[10])
         df_view1 = asarray_or_pandas(df)
-        df_view1.ix[10, "A"] = 101
+        df_view1.loc[10, "A"] = 101
         assert np.array_equal(df_view1.columns, ["A", "B", "C"])
         assert np.array_equal(df_view1.index, [10])
-        assert df.ix[10, "A"] == 101
+        assert df.loc[10, "A"] == 101
         df_copy = asarray_or_pandas(df, copy=True)
         assert np.array_equal(df_copy, df)
         assert np.array_equal(df_copy.columns, ["A", "B", "C"])
         assert np.array_equal(df_copy.index, [10])
-        df_copy.ix[10, "A"] = 100
+        df_copy.loc[10, "A"] = 100
         assert not np.array_equal(df_copy, df)
         df_converted = asarray_or_pandas(df, dtype=float)
         assert df_converted["A"].dtype == np.dtype(float)
@@ -202,7 +229,7 @@ def pandas_friendly_reshape(a, new_shape):
         return a
     if len(new_shape) == 1 and a.shape[1] == 1:
         if new_shape[0] != a.shape[0]:
-            raise ValueError, "arrays have incompatible sizes"
+            raise ValueError("arrays have incompatible sizes")
         return a[a.columns[0]]
     raise ValueError("cannot reshape a DataFrame with shape %s to shape %s"
                      % (a.shape, new_shape))
@@ -266,12 +293,12 @@ else: # pragma: no cover
 
 def wide_dtype_for(arr):
     arr = np.asarray(arr)
-    if (np.issubdtype(arr.dtype, np.integer)
-        or np.issubdtype(arr.dtype, np.floating)):
+    if (safe_issubdtype(arr.dtype, np.integer)
+        or safe_issubdtype(arr.dtype, np.floating)):
         return widest_float
-    elif np.issubdtype(arr.dtype, np.complexfloating):
+    elif safe_issubdtype(arr.dtype, np.complexfloating):
         return widest_complex
-    raise ValueError, "cannot widen a non-numeric type %r" % (arr.dtype,)
+    raise ValueError("cannot widen a non-numeric type %r" % (arr.dtype,))
 
 def widen(arr):
     return np.asarray(arr, dtype=wide_dtype_for(arr))
@@ -302,13 +329,14 @@ class PushbackAdapter(object):
             return self._pushed.pop()
         else:
             # May raise StopIteration
-            return self._it.next()
+            return six.advance_iterator(self._it)
+    __next__ = next
 
     def peek(self):
         try:
-            obj = self.next()
+            obj = six.advance_iterator(self)
         except StopIteration:
-            raise ValueError, "no more data"
+            raise ValueError("no more data")
         self.push_back(obj)
         return obj
 
@@ -323,10 +351,10 @@ class PushbackAdapter(object):
 def test_PushbackAdapter():
     it = PushbackAdapter(iter([1, 2, 3, 4]))
     assert it.has_more()
-    assert it.next() == 1
+    assert six.advance_iterator(it) == 1
     it.push_back(0)
-    assert it.next() == 0
-    assert it.next() == 2
+    assert six.advance_iterator(it) == 0
+    assert six.advance_iterator(it) == 2
     assert it.peek() == 3
     it.push_back(10)
     assert it.peek() == 10
@@ -356,7 +384,6 @@ def test_PushbackAdapter():
 # Pretty printer docs:
 #   http://ipython.org/ipython-doc/dev/api/generated/IPython.lib.pretty.html
 
-from cStringIO import StringIO
 class _MiniPPrinter(object):
     def __init__(self):
         self._out = StringIO()
@@ -549,3 +576,159 @@ def test_iterable():
     assert iterable({"a": 1})
     assert not iterable(1)
     assert not iterable(iterable)
+
+##### Handling Pandas's categorical stuff is horrible and hateful
+
+# Basically they decided that they didn't like how numpy does things, so their
+# categorical stuff is *kinda* like how numpy would do it (e.g. they have a
+# special ".dtype" attribute to mark categorical data), so by default you'll
+# find yourself using the same code paths to handle pandas categorical data
+# and other non-categorical data. BUT, all the idioms for detecting
+# categorical data blow up with errors if you try them with real numpy dtypes,
+# and all numpy's idioms for detecting non-categorical types blow up with
+# errors if you try them with pandas categorical stuff. So basically they have
+# just poisoned all code that touches dtypes; the old numpy stuff is unsafe,
+# and you must use special code like below.
+#
+# Also there are hoops to jump through to handle both the old style
+# (Categorical objects) and new-style (Series with dtype="category").
+
+# Needed to support pandas < 0.15
+def pandas_Categorical_from_codes(codes, categories):
+    assert have_pandas_categorical
+
+    # Old versions of pandas sometimes fail to coerce this to an array and
+    # just return it directly from .labels (?!).
+    codes = np.asarray(codes)
+    if hasattr(pandas.Categorical, "from_codes"):
+        return pandas.Categorical.from_codes(codes, categories)
+    else:
+        return pandas.Categorical(codes, categories)
+
+def test_pandas_Categorical_from_codes():
+    if not have_pandas_categorical:
+        return
+    c = pandas_Categorical_from_codes([1, 1, 0, -1], ["a", "b"])
+    assert np.all(np.asarray(c)[:-1] == ["b", "b", "a"])
+    assert np.isnan(np.asarray(c)[-1])
+
+# Needed to support pandas < 0.15
+def pandas_Categorical_categories(cat):
+    # In 0.15+, a categorical Series has a .cat attribute which is similar to
+    # a Categorical object, and Categorical objects are what have .categories
+    # and .codes attributes.
+    if hasattr(cat, "cat"):
+        cat = cat.cat
+    if hasattr(cat, "categories"):
+        return cat.categories
+    else:
+        return cat.levels
+
+# Needed to support pandas < 0.15
+def pandas_Categorical_codes(cat):
+    # In 0.15+, a categorical Series has a .cat attribute which is a
+    # Categorical object, and Categorical objects are what have .categories /
+    # .codes attributes.
+    if hasattr(cat, "cat"):
+        cat = cat.cat
+    if hasattr(cat, "codes"):
+        return cat.codes
+    else:
+        return cat.labels
+
+def test_pandas_Categorical_accessors():
+    if not have_pandas_categorical:
+        return
+    c = pandas_Categorical_from_codes([1, 1, 0, -1], ["a", "b"])
+    assert np.all(pandas_Categorical_categories(c) == ["a", "b"])
+    assert np.all(pandas_Categorical_codes(c) == [1, 1, 0, -1])
+
+    if have_pandas_categorical_dtype:
+        s = pandas.Series(c)
+        assert np.all(pandas_Categorical_categories(s) == ["a", "b"])
+        assert np.all(pandas_Categorical_codes(s) == [1, 1, 0, -1])
+
+# Needed to support pandas >= 0.15 (!)
+def safe_is_pandas_categorical_dtype(dt):
+    if not have_pandas_categorical_dtype:
+        return False
+    return _pandas_is_categorical_dtype(dt)
+
+# Needed to support pandas >= 0.15 (!)
+def safe_is_pandas_categorical(data):
+    if not have_pandas_categorical:
+        return False
+    if isinstance(data, pandas.Categorical):
+        return True
+    if hasattr(data, "dtype"):
+        return safe_is_pandas_categorical_dtype(data.dtype)
+    return False
+
+def test_safe_is_pandas_categorical():
+    assert not safe_is_pandas_categorical(np.arange(10))
+
+    if have_pandas_categorical:
+        c_obj = pandas.Categorical(["a", "b"])
+        assert safe_is_pandas_categorical(c_obj)
+
+    if have_pandas_categorical_dtype:
+        s_obj = pandas.Series(["a", "b"], dtype="category")
+        assert safe_is_pandas_categorical(s_obj)
+
+# Needed to support pandas >= 0.15 (!)
+# Calling np.issubdtype on a pandas categorical will blow up -- the officially
+# recommended solution is to replace every piece of code like
+#   np.issubdtype(foo.dtype, bool)
+# with code like
+#   isinstance(foo.dtype, np.dtype) and np.issubdtype(foo.dtype, bool)
+# or
+#   not pandas.is_categorical_dtype(foo.dtype) and issubdtype(foo.dtype, bool)
+# We do the latter (with extra hoops) because the isinstance check is not
+# safe. See
+#   https://github.com/pydata/pandas/issues/9581
+#   https://github.com/pydata/pandas/issues/9581#issuecomment-77099564
+def safe_issubdtype(dt1, dt2):
+    if safe_is_pandas_categorical_dtype(dt1):
+        return False
+    return np.issubdtype(dt1, dt2)
+
+def test_safe_issubdtype():
+    assert safe_issubdtype(int, np.integer)
+    assert safe_issubdtype(np.dtype(float), np.floating)
+    assert not safe_issubdtype(int, np.floating)
+    assert not safe_issubdtype(np.dtype(float), np.integer)
+
+    if have_pandas_categorical_dtype:
+        bad_dtype = pandas.Series(["a", "b"], dtype="category")
+        assert not safe_issubdtype(bad_dtype, np.integer)
+
+def no_pickling(*args, **kwargs):
+    raise NotImplementedError(
+        "Sorry, pickling not yet supported. "
+        "See https://github.com/pydata/patsy/issues/26 if you want to "
+        "help.")
+
+def assert_no_pickling(obj):
+    import pickle
+    from nose.tools import assert_raises
+    assert_raises(NotImplementedError, pickle.dumps, obj)
+
+# Use like:
+#   if safe_string_eq(constraints, "center"):
+#       ...
+# where 'constraints' might be a string or an array. (If it's an array, then
+# we can't use == becaues it might broadcast and ugh.)
+def safe_string_eq(obj, value):
+    if isinstance(obj, six.string_types):
+        return obj == value
+    else:
+        return False
+
+def test_safe_string_eq():
+    assert safe_string_eq("foo", "foo")
+    assert not safe_string_eq("foo", "bar")
+
+    if not six.PY3:
+        assert safe_string_eq(unicode("foo"), "foo")
+
+    assert not safe_string_eq(np.empty((2, 2)), "foo")
